/*
 * PingExampleAsync.java 
 *
 * Copyright (c) Verax Systems.
 * All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
package com.veraxsystems.icmp.ping.example;

import com.veraxsystems.icmp.jni.IcmpEchoResponse;
import com.veraxsystems.icmp.ping.Ping;
import com.veraxsystems.icmp.ping.PingResponseHandler;
import com.veraxsystems.icmp.ping.exception.PingErrorException;
import com.veraxsystems.icmp.ping.handlers.PingRequest;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.concurrent.CountDownLatch;

/**
 * The example program for {link Ping} class.
 * It accepts command line parameters and issues asynchronous set of
 * ICMP echo requests. Results of requests are displayed on the screen.
 */
public class PingExampleAsync
{
    
   static {
       com.veraxsystems.icmp.jni.IcmpJniBridge.loadNativeLibrary();       
   }
    
   private static CountDownLatch countDown = null;

   static class Handler implements PingResponseHandler
   {
      public synchronized void handleResponse(PingRequest request, IcmpEchoResponse response) 
      {
         System.out.println(request.getAddress().getHostAddress() + " from " + response.getSourceAddress().getHostAddress() + " response: " + response.getStatus() + " time:" + (response.getResponseTime() - request.getSendTime()));
         countDown.countDown();
      }
      
      public synchronized void handleTimeout(PingRequest request) 
      {
         System.out.println(request.getAddress().getHostAddress() + " timeout");
         countDown.countDown();
      }
      
      public synchronized void handleException(PingRequest request, Exception exception) 
      {
         exception.printStackTrace();
         System.out.println(request.getAddress().getHostAddress() + "ping error " + exception.toString());
         countDown.countDown();
      }
      
   }
 
   public static void main(String[] args) 
   {
      if (args.length >= 2)
      {
         try
         {
            InetAddress[] addresses = new Inet4Address[args.length - 1];
            int timeout = Integer.parseInt(args[0]);
            Handler handler = new Handler();
            int requestsNumber = 0;
            
            // Initialize Ping
            Ping.init();
            
            // Get names of hosts so it is easy to use CountDownLatch 
            for(int i = 1; i < args.length; i++)
            {
               try 
               {
                  addresses[requestsNumber] = Inet4Address.getByName(args[i]);
                  requestsNumber += 1;
               } 
               catch (UnknownHostException ex) 
               {
                  System.out.println("Unknown host: " + args[i]);
               }
            };
            
            // Initialize waiting counter so program can wait until all requests are handled
            countDown = new CountDownLatch(requestsNumber);
            try
            {
               // Send all requests
               for(int i = 0; i < requestsNumber; i++)
               {
                  Ping.echo(addresses[i], "DATA".getBytes(), timeout, handler);
               };
            }
            catch(PingErrorException ex)
            {
               // Fatal error occured in echo method - exit program
               System.out.println("Fatal error - terminating " + ex.toString());
            }
            try 
            {
               countDown.await();
            }
            catch (InterruptedException ex) 
            {
               System.out.println("Proram terminated- exiting");
            };
         } 
         catch (PingErrorException ex) 
         {
            System.out.println("Could not initialize Ping class: " + ex.toString());
         }
         finally
         {
            // Release Ping 
            Ping.release();
            
            // Release foces pending requests to fail - wait for them if necessary
            try 
            {
               if (countDown != null) countDown.await();
            }
            catch (InterruptedException ex) 
            {
               // Nothing to do - program is exiting
            };
         }
      }
      else
      {
         System.out.println("Usage:");
         System.out.println("     Ping <timeout> <address list>");
         System.out.println("");
      }
   }
}
