/*
 * Icmp.c
 *
 * Copyright (c) Verax Systems.
 * All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
//
//
// ICMP.C -- ICMP API using ICMP and RAW Sockets
// 
//

#include <stdio.h>
#include <winsock.h>
#include "Icmp.h"
#include "IcmpDebug.h"

/*
 *			I N _ C K S U M
 *
 * Checksum routine for Internet Protocol family headers (C Version)
 *
 */
static UINT16 in_cksum(u_short *addr, int len)
{
   register int nleft = len;
   register u_short *w = addr;
   register u_short answer;
   register int sum = 0;

   /*
    *  Our algorithm is simple, using a 32 bit accumulator (sum),
    *  we add sequential 16 bit words to it, and at the end, fold
    *  back all the carry bits from the top 16 bits into the lower
    *  16 bits.
    */
   while( nleft > 1 )  {
      sum += *w++;
      nleft -= 2;
   }

   /* mop up an odd byte, if necessary */
   if( nleft == 1 ) {
      u_short	u = 0;

      *(u_char *)(&u) = *(u_char *)w ;
      sum += u;
   }

   /*
    * add back carry outs from top 16 bits to low 16 bits
    */
   sum = (sum >> 16) + (sum & 0xffff);	/* add hi 16 to low 16 */
   sum += (sum >> 16);			/* add carry */
   answer = ~sum;				/* truncate to 16 bits */
   return (answer);
}



int IcmpInitialize(void)
{
   WORD wVersionRequested;
   WSADATA wsaData;
   int	result;

   wVersionRequested = MAKEWORD(2, 2);
   result = WSAStartup(wVersionRequested, &wsaData);
   IcmpOutputDebugString("IcmpInitialize: Initialization code %d", result);
   return result;
} /* IcmpInitialize */


void IcmpRelease(void)
{
   WSACleanup();
   IcmpOutputDebugString("IcmpRelease");
   return;
} /* IcmpRelease */


/*
 * Initialize RAW socket
 */
SOCKET IcmpCreateSocket(void)
{
   SOCKET rawSocket;

   IcmpOutputDebugString("IcmpCreateSocket IN\n");
   rawSocket = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
   if (rawSocket != INVALID_SOCKET)
      IcmpOutputDebugString("IcmpCreateSocket OUT: Socket %u created\n", rawSocket);
   else
      IcmpOutputDebugString("IcmpCreateSocket OUT: Socket creation error %d\n", WSAGetLastError());
   return rawSocket;
} /* IcmpCreateSocket */


/*
 * Close ICMP socket
 */
int IcmpCloseSocket(SOCKET rawSocket)
{
   int result;
   u_long iMode=1;

   IcmpOutputDebugString("IcmpCloseSocket IN: Socket %u\n", rawSocket);
   ioctlsocket(rawSocket,FIONBIO,&iMode);
   result = closesocket(rawSocket);
   if (result != SOCKET_ERROR)
      IcmpOutputDebugString("IcmpCloseSocket OUT: Socket %u closed\n", rawSocket);
   else
      IcmpOutputDebugString("IcmpCloseSocket OUT: Socket %u close error %d\n", rawSocket, WSAGetLastError());
   return result;
} /* IcmpCloseSocket */


/*
 * Get last error reported for given socket.
 */
int IcmpGetSocketError(SOCKET rawSocket)
{
   int errorCode;
   int errorCodeLen = sizeof(errorCode);
   int result;

   result = getsockopt(rawSocket, SOL_SOCKET, SO_ERROR, (char *)&errorCode, &errorCodeLen);
   if (result == SOCKET_ERROR)
   {
      IcmpOutputDebugString("IcmpGetSocketError: Socket %u getsockop error %d", rawSocket, result);
      errorCode = WSAEINVAL;
   };
   return errorCode;
} /* IcmpGetSocketError */


/*
 * Get last error
 */
int IcmpGetLastError(void)
{
   IcmpOutputDebugString("IcmpGetLastError: Get last error %d", WSAGetLastError());
   return WSAGetLastError();
} /* IcmpGetLastError */


/*
 * Send ICMP Echo request. Request is sent asynchronously and function returns code of operation 
 * according to WIN32 @see sendto implementation.
 */
int IcmpSendEcho(SOCKET rawSocket, const INT8 *destRawAddress, UINT16 identifier, UINT16 sequenceNumber, void *dataBuffer, size_t dataLength)
{
   // build the packet
   IcmpMessage echoRequest;
   struct sockaddr_in dstAddr;
   size_t requestLength;
   int result;
 
   IcmpOutputDebugString("IcmpSendEcho IN: Socket %u\n", rawSocket);
   memset(&dstAddr, 0, sizeof(dstAddr));
   if (dataLength <= ICMP_ECHO_MAX_DATA_LEN  && ! IsBadReadPtr(destRawAddress, sizeof(dstAddr.sin_addr.S_un.S_addr)) && ! IsBadReadPtr(dataBuffer, dataLength))
   {
      /* Copy raw IP addres (4 bytes) to appropriate sockaddr structure */
      dstAddr.sin_family = AF_INET;
      memcpy(&dstAddr.sin_addr.S_un.S_addr,  destRawAddress, sizeof(dstAddr.sin_addr.S_un.S_addr));
      IcmpOutputDebugString("IcmpSendEcho: Socket %u IP=%u.%u.%u.%u\n", rawSocket, (UINT8)(destRawAddress[0]), (UINT8)(destRawAddress[1]), (UINT8)(destRawAddress[2]), (UINT8)(destRawAddress[3]));

      /* Build ICMP Echo request */
      memset(&echoRequest, 0, sizeof (echoRequest));
      echoRequest.header.type = ICMP_ECHOREQEST;                  // RFC 792 type: 8 = ECHO_REQUEST
      echoRequest.header.code = 0;                                // RFC 792 type: code
      echoRequest.header.quench.echo.id = htons(identifier);      // RFC 792 type: identification
      echoRequest.header.quench.echo.seq = htons(sequenceNumber); // RFC 792 type: sequenceNumber
      memcpy(echoRequest.data.buffer, dataBuffer, dataLength);
      requestLength = sizeof(IcmpMessage) - ICMP_ECHO_MAX_DATA_LEN + dataLength;
      echoRequest.header.checksum = in_cksum((u_short *)&echoRequest, requestLength);

      result = sendto(rawSocket, (const char*) &echoRequest, (int)requestLength, 0, (struct sockaddr *)&dstAddr, sizeof(struct sockaddr_in)); 
   }
   else
   {
      WSASetLastError(WSAEINVAL);
      result = SOCKET_ERROR;
   };
   if (result != SOCKET_ERROR)
      IcmpOutputDebugString("IcmpSendEcho OUT: Socket %u sendto OK", rawSocket, dstAddr.sin_addr.S_un.S_addr);
   else
      IcmpOutputDebugString("IcmpSendEcho OUT: Socket %u sendto failed %d", rawSocket, WSAGetLastError());
   return result; 
} /* IcmpSendEchoRequest */


/*
 * Waits until either data is available on socket or timeout occurs.
 * Additional checking is made so negative timeout parameter forces function to behave like
 * when timeout occurs.
 */
int IcmpWait(SOCKET rawSocket, long timeout)
{
   struct timeval timeoutVal;
   fd_set readFd;
   int result;

   IcmpOutputDebugString("IcmpWait IN: Socket %u timeout %ld", rawSocket, timeout);
   if (timeout >= 0)
   {
      timeoutVal.tv_sec = timeout / 1000;
      timeoutVal.tv_usec = 1000 * (timeout % 1000);
      
      readFd.fd_count = 1;
      readFd.fd_array[0] = rawSocket;
      
      result = select(1, &readFd, NULL, NULL, &timeoutVal);
   }
   else result = 0;
   if (result != SOCKET_ERROR)
      IcmpOutputDebugString("IcmpWait OUT: Socket %u select OK result = %d", rawSocket, result);
   else
      IcmpOutputDebugString("IcmpWait OUT: Socket %u select failed %d", rawSocket, WSAGetLastError());
   return result;
} /* IcmpWait */


/*
 * Reads the message from raw socket. 
 */
int IcmpReadMessage(SOCKET rawSocket, IcmpResponseMessage *response)
{
   struct sockaddr_in sourceAddress;
   int sourceAddLen = sizeof(struct sockaddr_in);
   int result;
   
   IcmpOutputDebugString("IcmpReadMessage IN: Socket %u", rawSocket);
   if (! IsBadReadPtr(response, sizeof(IcmpResponseMessage)))
   {
      result = recvfrom(rawSocket, (char *)response, sizeof(IcmpResponseMessage), 0, (struct sockaddr *)&sourceAddress, &sourceAddLen);
   }
   else
   {
      WSASetLastError(WSAEINVAL);
      result = SOCKET_ERROR;
   };
   if (result != SOCKET_ERROR)
      IcmpOutputDebugString("IcmpReadMessage OUT: Socket %u recvfrom OK", rawSocket);
   else
      IcmpOutputDebugString("IcmpReadMessage OUT: Socket %u recvrom failed %d", rawSocket, WSAGetLastError());
   return result;
} /* IcmpReadMessage */


/*
 * Decode ICMP response message to ICMP echo components.
 */
BOOL IcmpDecodeResponseMessage(const IcmpResponseMessage *message,  UINT8 *srcAddress, UINT8 *dstAddress, UINT8 *type, UINT8 *code, UINT16 *identifier, UINT16 *sequenceNumber, void *dataBuffer, size_t dataBuffferLength, size_t *dataLength)
{
   size_t len;
   BOOL result;
   
   IcmpOutputDebugString("IcmpDecodeResponseMessage IN:\n");
   if (! IsBadReadPtr(message, sizeof(IcmpResponseMessage)) && ! IsBadWritePtr(srcAddress, 4 * sizeof(UINT8)) && ! IsBadWritePtr(type, sizeof(INT8)) && ! IsBadWritePtr(code, sizeof(INT8)) && ! IsBadWritePtr(identifier, sizeof(UINT16)) && ! IsBadWritePtr(sequenceNumber, sizeof(UINT16))&& ! IsBadWritePtr(dataLength, sizeof(size_t)))
   {
      IcmpOutputDebugString("IcmpDecodeResponseMessage: Decode IP protocol %d\n", message->ipHeader.protocol);
      if (message->ipHeader.protocol == IP_PROTOCOL_ICMP)
      {
         IcmpOutputDebugString("ICMP: Decode message type %d\n", message->icmp.header.type);
         switch(message->icmp.header.type)
         {
            case ICMP_ECHOREPLY :
                len = ntohs(message->ipHeader.length) - ICMP_PAYLOAD_MIN_LEN - IP_HEADER_LEN;
                if (len <= dataBuffferLength && ! IsBadWritePtr(dataBuffer, len))
                {
                   *type = message->icmp.header.type;
                   *code = message->icmp.header.code;
                   *identifier = ntohs(message->icmp.header.quench.echo.id);
                   *sequenceNumber = ntohs(message->icmp.header.quench.echo.seq);
                   memcpy(dataBuffer, message->icmp.data.buffer, len);
                   *dataLength = len;
                   memcpy(srcAddress, &(message->ipHeader.iaddrSrc.S_un.S_addr), sizeof(message->ipHeader.iaddrSrc.S_un.S_addr));
                   /* There is no echo destination address in response but we assume the echo response is sent from echo request destination */
                   memcpy(dstAddress, &(message->ipHeader.iaddrSrc.S_un.S_addr), sizeof(message->ipHeader.iaddrSrc.S_un.S_addr));
                   result = TRUE;
                }
                else
                {
                   WSASetLastError(WSAEFAULT);
                   result = FALSE;
                };
                break;
            case ICMP_DESTINATIONUNREACHABLE:
            case ICMP_SOURCEQUENCH:
            case ICMP_REDIRECT:
            case ICMP_ECHOREQEST:
            case ICMP_TIMEEXCEEDED:
            case ICMP_PARAMETERPROBLEM:
                *type = message->icmp.header.type;
                *code = message->icmp.header.code;
                *identifier = ntohs(message->icmp.data.packet.icmp.quench.echo.id);
                *sequenceNumber = ntohs(message->icmp.data.packet.icmp.quench.echo.seq);
                *dataLength = 0;
                memcpy(srcAddress, &(message->ipHeader.iaddrSrc.S_un.S_addr), sizeof(message->ipHeader.iaddrSrc.S_un.S_addr));
                memcpy(dstAddress, &(message->icmp.data.packet.ipHeader.iaddrDst.S_un.S_addr), sizeof(message->ipHeader.iaddrSrc.S_un.S_addr));
                result = TRUE;
                break;
            default:
                WSASetLastError(WSAEPROTONOSUPPORT);
                result = FALSE;
                break;
         };
      }
      else
      {
         WSASetLastError(WSAEPROTONOSUPPORT);
         result = FALSE;
      };
   }
   else 
   {
      WSASetLastError(WSAEINVAL);
      result = FALSE;
   };
   if (result == TRUE)
      IcmpOutputDebugString("IcmpDecodeResponseMessage OUT: Decode message OK");
   else
      IcmpOutputDebugString("IcmpDecodeResponseMessage OUT: Decode message failed %d", result);
   return result;
} /* IcmpDecodeResponseMessage */


/*
 * Get system error string assiciated with error code.
 */
LPTSTR IcmpGetErrorString(int errorCode)
{
    LPTSTR errorStr = NULL;

    FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_ARGUMENT_ARRAY | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, errorCode, 0, (LPSTR)&errorStr, 0, NULL);
    return errorStr;
} /* IcmpGetErrorString */


/*
 * Get system error string associated with error code.
 */
void IcmpReleaseErrorString(LPTSTR errorStr)
{
   if (errorStr != NULL) LocalFree(errorStr);
   return;
} /* IcmpReleaseErrorString */


