/*
 * Icmp.h
 *
 * Copyright (c) Verax Systems.
 * All rights reserved.
 *
 * This software is furnished under a license. Use, duplication,
 * disclosure and all other uses are restricted to the rights
 * specified in the written license agreement.
 */
/**
 * @file Icmp.h
 *
 * @brief ICMP API: basic types and functions. 
 *
 * File contains definitions of data structures and functions sufficient to
 * send ICMP packet, receive response and proper error handling.
 * 
 */


#ifndef __ICMP_H__
#define __ICMP_H__

#include <windows.h>

/* Pack data structures so no data padding is used in messages/packets definitions */
#pragma pack(1)

#define IP_PROTOCOL_ICMP             1

#define ICMP_ECHOREPLY               0
#define ICMP_DESTINATIONUNREACHABLE  3
#define ICMP_SOURCEQUENCH            4
#define ICMP_REDIRECT                5
#define ICMP_ECHOREQEST              8
#define ICMP_TIMEEXCEEDED           11
#define ICMP_PARAMETERPROBLEM       12

/**
 * @brief The length of IP header expressed in bytes 
 */
#define IP_HEADER_LEN        20     

/**
 * @brief The minimum length of ICMP payload message - such message may be followed by additional data 
 */
#define ICMP_PAYLOAD_MIN_LEN 8


/**
 * @brief The length of data attached to ICMP messages in case the ICMP message reports error condition 
 */
#define ICMP_ERROR_MESSAGE_LEN (IP_HEADER_LEN + ICMP_PAYLOAD_MIN_LEN)

/**
 * @brief The maximum size of IP packet 
 */
#define IP_MESSAGE_MAX_LEN   0xFFFF 

/**
 * @brief The maximum size of ICMP echo data content. 
 */
#define ICMP_ECHO_MAX_DATA_LEN  (IP_MESSAGE_MAX_LEN - ICMP_PAYLOAD_MIN_LEN - IP_HEADER_LEN)



/**
 * @brief IP Header as defined in RFC 791
 */
typedef
   struct 
   {
      UINT8  versionIhl;                  /**< @brief Version, IHL (compressed on one byte) */
      UINT8  tos;                         /**< @brief Type Of Service */
      UINT16 length;                      /**< @brief Total Length */
      UINT16 id;                          /**< @brief Identification */
      UINT16 flagsOffset;                 /**< @brief Flags, Fragment Offset (compressed on two bytes) */
      UINT8  ttl;                         /**< @brief Time To Live */
      UINT8  protocol;                    /**< @brief Protocol */
      UINT16 checksum;                    /**< @brief Header checksum */
      struct in_addr   iaddrSrc;                    /**< @brief Source address */
      struct in_addr   iaddrDst;                    /**< @brief Destination address */
   } IpHeader;



/**
 * @brief ICMP Redirect quench as defined in RFC 792
 */
typedef 
   struct
   {
      struct in_addr gatewayAddr;         /**< @brief  Address of the gateway */
   } IcmpRedirectQuench;


/**
 * @brief ICMP Echo Request or Echo Response quench as defined in RFC 792
 */
typedef 
   struct
   {
      UINT16 id;                          /**< @brief  Identifier */
      UINT16 seq;                         /**< @brief  Sequence Number */
   } IcmpEchoQuench;


typedef
   union
   {
      IcmpRedirectQuench redirect;
      IcmpEchoQuench     echo;
   } IcmpQuench;

/**
 * @brief ICMP Message common header content as defined in RFC 792
 */
typedef 
   struct
   {
      UINT8      type;                        /**< @brief  Type */
      UINT8      code;                        /**< @brief  Code */
      UINT16     checksum;                    /**< @brief  Checksum */
      IcmpQuench quench;
   } IcmpHeader;


/**
 * @brief ICMP response data.
 * Referring to ICMP specification, the ICMP response message contains additional data. This data
 * includes IP message header and 16 bytes of IP message data. In this library, response
 * always contains 16 bytes of ICMP message that is IcmpHeader structure.
 */
typedef
   struct
   {
      IpHeader     ipHeader;                    /**< @brief IP header */
      IcmpHeader   icmp;                        /**< @brief ICMP payload message */
   } IcmpRawMessage;


/**
 * @brief ICMP response data.
 * The data may be at most ICMP_ECHO_MAX_DATA_LEN long.
 * In this library the data content of ICMP message contains either raw data (this may be fetched
 * from buffet field) or structured ICMP raw message (this includes IP message header).
 */
typedef
   union
   {
      IcmpRawMessage packet;
      UINT8          buffer[ICMP_ECHO_MAX_DATA_LEN];/**< @brief  ICMP data placeholder (the real data size is calculated from IP header length) */
   } IcmpData;


/**
 * @brief ICMP Message as defined in RFC 792.
 * The message contains ICMP header where type of request and request specific
 * attributes (quench) are set and optional message data. 
 */
typedef
   struct
   {
      IcmpHeader header;
      IcmpData   data;
   } IcmpMessage;


/**
 * @brief ICMP response message. It contains IP header and ICMP message as defined in RFC 792.
 * Such structure is defined because this is the content of ICMP response received from WIN32
 * raw socket.
 */
typedef
   struct
   {
      IpHeader    ipHeader;                    /**< @brief IP header */
      IcmpMessage icmp;
   } IcmpResponseMessage;


#pragma pack()

/**
 * IcmpSocket
 */

/**
 * @brief Initialize ICMP library
 *
 * Function initializes ICMP library, especially sockets by loading ws2.dll.
 * It  returns code of operation that can be found in documentation of Win32 WSAStartup function.
 * In order to release resources allocated by this function, the IcmpRelease() must be executed.
 *
 * @return code of operation, 0 means success
 */
int IcmpInitialize(void);


/**
 * @brief Release resources allocated by IcmpInitialize() function.
  */
void IcmpRelease(void);


/**
 * @brief Create socket for ICMP communictario
 *
 * Create socket in raw mode so it may be used for ICMP communication.
 * Remember to close socket using IcmpCreateSocket() function in order to release system resources.
 *
 * @return handle to socket od SOCKET_ERROR in case of error
 */
SOCKET IcmpCreateSocket(void);


/**
 * @brief Close socket opened by IcmpCreateSocket()
 *
 * In order to release system resources this function must be executed for each socket created by IcmpCreateSocket
 * function. It is thread safe an may be run from different thread than socket was created.
 *
 * @param socket socket to be closed
 * @return SOCKET_ERROR in case of error, other value in success
 */
int IcmpCloseSocket(SOCKET socket);


/**
 * @brief Wait until socket has data to read.
 *
 * Function implements behavior of <i>select</i> functionality for receiving data from socket.
 * It waits until some data is ready to read from socket or given timeout occurs.
 *
 * @param socket socket to wait for data
 * @param timeout maximum time (in milliseconds) to wait for data
 * @return 0 on timeout, 1 when data is ready or SOCKET_ERROR in case of error
 * @remarks Function terminates if socket is closed (this may be don in other thread)
 */
int IcmpWait(SOCKET socket, long timeout);


/**
 * @brief Send ICMP EchoRequest
 *
 * Function sends ICMP echo message to given destination host. The ICMP message is constructed
 * internally and is filled by given identifier, sequenceNumber and data (see IcmpMessage structure)
 * Request is sent asynchronously and function returns code of operation 
 * according to WIN32 sendto implementation.
 *
 * @param socket raw socket the message will be sent through
 * @param destAddress address to send message to
 * @param identifier identifier of ICMP message
 * @param sequenceNumber sequence number of ICMP message
 * @param dataBuffer data to be send in ICMP message
 * @param dataLength length of data
 * @return size of sent message of SOCKET_ERROR in case of error
 */
int IcmpSendEcho(SOCKET socket, const INT8 *destAddress, UINT16 identifier, UINT16 sequenceNumber, void *dataBuffer, size_t dataLength);


/**
 * @brief Read message from socket assuming received message has ICMP structure.
 *
 *  Function reads data from raw socket. It is assumed that it is ICMP message, but function does
 * not verify this.
 * Function does not change socket mode (blocking/not blocking) so application must
 * properly handle socket behavior.
 * When function file detailed error code may be obtained by IcmpGetLastError() function. 
 * @param socket socket to read message from    
 * @param response the message received
 * @return size of received message or SOCKET_ERROR in case of error
 */
int IcmpReadMessage(SOCKET socket, IcmpResponseMessage *response);


/**
 * @brief Decode ICMP response message.
 *
 * Function takes IcmpResponseMessage and fetches its ICMP attributes. In case message cannot be recognized,
 * the FALSE value is returned.
 *
 * @param message message to be decoded
 * @param srcAddress message sender address 
 * @param dstAddress ICMP EchoRequest destination address
 * @param type type of ISMP message
 * @param code the code related to message type (see RFC 791)
 * @param identifier message ID
 * @param sequenceNumber message sequence number
 * @param dataBuffer data buffer to store decoded ICMP EchoResponse data
 * @param dataBuffferLength maximum length of data buffer
 * @param dataLength the real length of received data
 * @return The return value indicates status of operation. It is TRUE on success or FALSE in case of failure. More detailed operation status may be obtained
 */
BOOL   IcmpDecodeResponseMessage(const IcmpResponseMessage *message,  UINT8 *srcAddress, UINT8 *dstAddress, UINT8 *type, UINT8 *code, UINT16 *identifier, UINT16 *sequenceNumber, void *dataBuffer, size_t dataBuffferLength, size_t *dataLength);


/**
 * @brief Get code of error reported by socket
 *
 * The error code is retrieved from socket using <i>getsockopt</i> function
 *
 * @param socket socket to read error from
 * @return error code
 */
int    IcmpGetSocketError(SOCKET socket);


/**
 * @brief Get WIN32 API error code
 *
 * Function returns WIN32 API error code that was set by WIN32 API function.
 *
 * @return Code of error
 */
int    IcmpGetLastError(void);


/**
 * @brief Get error string
 *
 * Function returns string describing given error code. String content is handled
 * by Windows Operation System
 * Function allocates resources for returned string so it must be released by <i>IcmpReleaseErrorString</i>
 * function.
 *
 * @param errorCode code of error to be described
 * @return string describing error code
 */
LPTSTR IcmpGetErrorString(int errorCode);


/**
 * @brief Release string allocated by <i>IcmpGetErrorString</i> function
 *
 * @param errorStr string allocated by <i>IcmpGetErrorString</i> function
 */
void   IcmpReleaseErrorString(LPTSTR errorStr);

#endif
